function createArray(length) {
    var arr, i, args;
    arr = [length || 0];
    i = length;

    if (arguments.length > 1) {
        args = Array.prototype.slice.call(arguments, 1);
        while (i--) {
            arr[length - 1 - i] = createArray.apply(this, args);
        }
    }

    return arr;
}

function pdnaHash(pb, width, height) {
    var r, c, i, j, len, hashLength, hashValue, sumImage, intensity, quadSize, numBins, patchSize, patchSampleDistH, patchSampleDistV, box, boxMult, patch, sigma0, sigma1, sigma2, sigma, scale, scalar, x0, y0, x1, y1, x1in, y1in, boxAvg, ix, iy, xd, yd, val, histogram, rbin, cbin, rd, cd, rb, cb, histIdx, featureOrien, offset1, offset2, histInitId, diffr, diffc, rect0, rect1, rect2, rect3, rf, cf, rd1, cd1, del, iter, found, maxThres, k, denom;
    r = 0;
    c = 0;
    i = 0;
    j = 0;
    len = 0;
    sumImage = [width * height];
    hashValue = [];

    // If image is too small return
    if (width < 50 || height < 50) {
        return null;
    }

    // First step is to create an integral image used for quick downsampling

    // Handle border cases
    intensity = pb[len++];
    intensity += pb[len++];
    intensity += pb[len++];
    len++;
    intensity /= 3.0;

    sumImage[0] = intensity;

    for (c = 1; c < width; c++) {
        intensity = pb[len++];
        intensity += pb[len++];
        intensity += pb[len++];
        len++;
        intensity /= 3.0;
        sumImage[c] = intensity + sumImage[c - 1];
    }

    for (r = 1; r < height; r++) {
        intensity = pb[len++];
        intensity += pb[len++];
        intensity += pb[len++];
        len++;
        intensity /= 3.0;
        sumImage[r * width] = intensity + sumImage[(r - 1) * width];

        for (c = 1; c < width; c++) {
            // Compute rest of pixel values after border cases are handled
            intensity = pb[len++];
            intensity += pb[len++];
            intensity += pb[len++];
            len++;
            intensity /= 3.0;
            sumImage[r * width + c] = intensity + sumImage[(r - 1) * width + c] + sumImage[r * width + c - 1] - sumImage[(r - 1) * width + c - 1];
        }
    }

    // Downsample the image into a (6*quadSize + 2) x (6*quadSize + 2) patch
    quadSize = 4;
    numBins = 6;
    patchSize = numBins * quadSize + 2;
    patchSampleDistH = width / (patchSize + 2);
    patchSampleDistV = height / (patchSize + 2);
    box = createArray(3, 4, 2);
    boxMult = createArray(3, 4);
    patch = [patchSize * patchSize];
    sigma0 = 0.5 * 0.8;
    sigma1 = 0.8;
    sigma2 = 2.0 * 0.8;

    // Approximate a Gaussian blur using three fast to compute box filters
    for (i = 0; i < 3; i++) {
        sigma = 0.0;
        scale = 0.0;

        if (i === 0) {
            sigma = sigma0;
            scale = Math.exp(-(sigma0 * sigma0) / 2.0) - Math.exp(-(sigma1 * sigma1) / 2.0);
        }

        if (i === 1) {
            sigma = sigma1;
            scale = Math.exp(-(sigma1 * sigma1) / 2.0) - Math.exp(-(sigma2 * sigma2) / 2.0);
        }

        if (i === 2) {
            sigma = sigma2;
            scale = Math.exp(-(sigma2 * sigma2) / 2.0);
        }

        j = 0;
        for (r = -1; r <= 1; r += 2) {
            for (c = -1; c <= 1; c += 2) {
                box[i][j][0] = c * patchSampleDistH * sigma / 2.0;
                box[i][j][1] = r * patchSampleDistV * sigma / 2.0;
                boxMult[i][j] = scale;
                j++;
            }
        }

        box[i][0][0]--;
        box[i][0][1]--;
        box[i][1][1]--;
        box[i][2][0]--;

        boxMult[i][1] = -boxMult[i][1];
        boxMult[i][2] = -boxMult[i][2];
    }

    scalar = patchSampleDistH * patchSampleDistV * sigma0 * sigma0 * boxMult[0][0] + patchSampleDistH * patchSampleDistV * sigma1 * sigma1 * boxMult[1][0] + patchSampleDistH * patchSampleDistV * sigma2 * sigma2 * boxMult[2][0];

    // Compute patch values
    for (r = 0; r < patchSize; r++) {
        for (c = 0; c < patchSize; c++) {
            x0 = (c + 1.5) * patchSampleDistH;
            y0 = (r + 1.5) * patchSampleDistV;

            patch[r * patchSize + c] = 0.0;

            for (i = 0; i < 3; i++) {
                boxAvg = 0.0;

                for (j = 0; j < 4; j++) {
                    x1 = x0 + box[i][j][0];
                    y1 = y0 + box[i][j][1];

                    // make sure sample points are within the image
                    // boundaries
                    x1in = Math.min(width - 2.0,
                            Math.max(0.0, x1));
                    y1in = Math.min(height - 2.0,
                            Math.max(0.0, y1));

                    // Bilinearly interpolate value
                    ix = parseInt(x1in);
                    iy = parseInt(y1in);

                    xd = x1in - ix;
                    yd = y1in - iy;

                    val = 0.0;

                    val += (1.0 - xd) * (1.0 - yd) * sumImage[iy * width + ix];
                    val += (1.0 - xd) * (yd) * sumImage[(iy + 1) * width + ix];
                    val += xd * (1.0 - yd) * sumImage[iy * width + ix + 1];
                    val += xd * yd * sumImage[(iy + 1) * width + (ix + 1)];

                    boxAvg += val * boxMult[i][j];
                }

                // Add box filter response to patch value
                patch[r * patchSize + c] += boxAvg;
            }

            patch[r * patchSize + c] /= scalar;

        }
    }

    // Compute edge histograms for fuzzy hash values
    histogram = [numBins * numBins * 4];
    featureOrien = 4;
    offset1 = numBins * featureOrien;
    offset2 = featureOrien;

    hashLength = numBins * numBins * featureOrien;

    // Initialize histogram values
    for (histInitId = 0; histInitId < numBins * numBins * 4; histInitId++) {
        histogram[histInitId] = 0;
    }

    // Create hash from rectified horizontal and vertical gradients
    for (rbin = 0; rbin < numBins; rbin++) {
        for (cbin = 0; cbin < numBins; cbin++) {
            for (rd = 0; rd < quadSize; rd++) {
                for (cd = 0; cd < quadSize; cd++) {
                    r = rbin * quadSize + rd + 1;
                    c = cbin * quadSize + cd + 1;

                    diffc = patch[r * patchSize + c - 1];
                    diffc -= patch[r * patchSize + c + 1];
                    diffr = patch[(r - 1) * patchSize + c];
                    diffr -= patch[(r + 1) * patchSize + c];

                    if (diffc > 0.0) {
                        rect0 = diffc;
                        rect1 = 0.0;
                    } else {
                        rect0 = 0.0;
                        rect1 = -diffc;
                    }

                    if (diffr > 0.0) {
                        rect2 = diffr;
                        rect3 = 0.0;
                    } else {
                        rect2 = 0.0;
                        rect3 = -diffr;
                    }

                    rf = r - 2.5;
                    cf = c - 2.5;

                    rf /= 4;
                    cf /= 4;

                    rb = parseInt(Math.floor(rf));
                    cb = parseInt(Math.floor(cf));

                    rd1 = rf - rb;
                    cd1 = cf - cb;

                    // Add gradients to histogram using bilinear
                    // interpolation
                    if (rb >= 0 && cb >= 0) {
                        del = (1.0 - rd1) * (1.0 - cd1);
                        histIdx = rb * offset1 + cb * offset2;
                        histogram[histIdx] += rect0 * del;
                        histogram[histIdx + 1] += rect1 * del;
                        histogram[histIdx + 2] += rect2 * del;
                        histogram[histIdx + 3] += rect3 * del;
                    }

                    if (rb >= 0 && cb < numBins - 1) {
                        del = (1.0 - rd1) * (cd1);
                        histIdx = rb * offset1 + (cb + 1) * offset2;
                        histogram[histIdx] += rect0 * del;
                        histogram[histIdx + 1] += rect1 * del;
                        histogram[histIdx + 2] += rect2 * del;
                        histogram[histIdx + 3] += rect3 * del;
                    }

                    if (rb < numBins - 1 && cb >= 0) {
                        del = (rd1) * (1.0 - cd1);
                        histIdx = (rb + 1) * offset1 + cb * offset2;
                        histogram[histIdx] += rect0 * del;
                        histogram[histIdx + 1] += rect1 * del;
                        histogram[histIdx + 2] += rect2 * del;
                        histogram[histIdx + 3] += rect3 * del;
                    }

                    if (rb < numBins - 1 && cb < numBins - 1) {
                        del = (rd1) * (cd1);
                        histIdx = (rb + 1) * offset1 + (cb + 1) * offset2;
                        histogram[histIdx] += rect0 * del;
                        histogram[histIdx + 1] += rect1 * del;
                        histogram[histIdx + 2] += rect2 * del;
                        histogram[histIdx + 3] += rect3 * del;
                    }
                }
            }
        }
    }

    // Make sure no value in the histogram is greater than 0.25. This
    // results is a better/more uniform distribution of hash values.
    iter = 0;
    found = true;
    maxThres = 0.25;

    while (iter < 10 && found) {
        iter++;
        denom = 0.00000001;

        for (rb = 0; rb < numBins; rb++) {
            for (cb = 0; cb < numBins; cb++) {
                for (i = 0; i < featureOrien; i++) {
                    histIdx = rb * offset1 + cb * offset2 + i;
                    denom += histogram[histIdx] * histogram[histIdx];
                }
            }
        }

        // If denom is zero return no hash. This should only happen if the
        // image is flat.
        if (denom <= 0.0) {
            return null;
        }

        denom = Math.sqrt(denom);

        found = false;

        for (rb = 0; rb < numBins; rb++) {
            for (cb = 0; cb < numBins; cb++) {
                for (i = 0; i < featureOrien; i++) {
                    histIdx = rb * offset1 + cb * offset2 + i;
                    histogram[histIdx] /= denom;

                    if (histogram[histIdx] >= maxThres && iter < 10) {
                        histogram[histIdx] = maxThres;
                        found = true;
                    }
                }
            }
        }

    }

    // Create final array to return hash values
    k = 0;
    for (rb = 0; rb < numBins; rb++) {
        for (cb = 0; cb < numBins; cb++) {
            for (i = 0; i < featureOrien; i++) {
                hashValue[k] = parseInt(Math.min(255.0, 256.0 * histogram[rb * offset1 + cb * offset2 + i] / maxThres));
                k++;
            }
        }
    }

    return hashValue;
}

function toHex(hash) {
    var hex, hexHash;
    hexHash = '';
    for (i = 0; i < 144; i++) {
        hex = hash[i].toString(16);
        hexHash += (hex.length === 1 ? '0' + hex : hex);
    }
    return hexHash.toUpperCase();
}

function distance(f, s) {
    var d, i;
    d = 0;

    for (i = 0; i < 144; i++) {
        d += (f[i] - s[i]) * (f[i] - s[i]);
    }
    return d;
}